#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3
#pragma version=4.13
#pragma IgorVersion=6
#pragma ModuleName=BaselineSpline

#include <Readback ModifyStr>

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 7 or later and have Project Updater 
// installed, this package can check periodically for new releases. 
// http: www.igorexchange.com/project/Updater
static constant kCHECK_FOR_UPDATES=1  // set to 1 to check for updates on init

static function CheckForUpdates()
	if(kCHECK_FOR_UPDATES)
		#if (exists("Updater#UpdateCheck")==6)
			return Updater#UpdateCheck("fileloc:"+FunctionPath(""))
		#endif
	endif
	return 0
end
// --------------------------------------------------------------------

// Written by Tony Withers, https://www.wavemetrics.com/user/tony
// I would be very happy to hear from you if you find this package 
// useful, or if you have any suggestions for improvement. 

// would be nice if it worked with axes other than left/bottom

// How to use Baseline Spline Fit

// A 1d wave must be plotted in a graph window before selecting Spline 
// Baseline from the macros menu. The wave must be plotted on the default
// bottom and left axes. A control panel will be added to the left of the
// graph window. When 'edit mode' is selected you can drag the nodes. The
// spline updates as the nodes are repositioned. When in edit mode you 
// can add a new node at any position on the graph by positioning the 
// mouse cursor and clicking while holding down the control key. Nodes 
// may be deleted by clicking them with the option key held down. 
// Clicking anywhere in the graph window with the shift key held down 
// will toggle between node editing mode and 'normal' mode where you can 
// interact with the graph as usual. Cubic and smoothing splines are 
// calculated by Igor's Interpolate2 operation; execute 'DisplayHelpTopic
// Interpolate2' for details of the methods of spline interpolation.

// Panel controls

// Data wave: select a trace for baseline correction. When switching 
// between traces the nodes will retain their X-positions but will be 
// placed on the selected data wave at those positions (unless they lie 
// outside of the range of the data wave).

// Akima: fit an Akima style spline through the nodes. Akima splines can 
// have sharper bends than cubic splines.

// PCHIP: fit a cubic Hermite spline through the nodes. This spline type 
// doesn't overshoot local maxima or minima in node sequence. First 
// derivative is continuous through the nodes, but second derivative may 
// be discontinuous.

// Cubic: fit a 'normal' cubic spline through the nodes. Second 
// derivative is continuous though the nodes.

// Ends: At the ends of the spline where the cubic function is not fully 
// constrained select a method to add a constraint. 1st deriv.: match 1st
// derivative; natural: match 2nd derivative. Affects only cubic spline 
// type.

// Linear: connect the nodes with linear segments.

// Smoothing: fit a smoothing spline, which does not have to pass exactly
// through the nodes. The degree of smoothing is adjustable. 

// Edit mode: toggle between node editing and normal modes

// Reset: set node positions by spacing a set number of nodes at regular 
// intervals along the section of the data wave displayed in the graph 
// window. No nodes will be set beyond the range of the graph axes.  

// Load: reload the node positions from a previous fit.

// Subtract: creates a copy of the baseline and a baseline-subtracted 
// output spectrum in the current data folder and appends them to the 
// plot; saves the node positions in a 2d wave in the current data 
// folder.


// Version history

// 4.14 corrects the SBL_EndDerivativePCHIP function for PCHIP splines.
// Previous versions may in some circumstances extrapolate beyond the end 
// nodes in an unreasonable way.
// 4.11 2/26/19 Added headers for Project Updater, fixed wave indexing bug 
// by switching to double precision precision node waves
// 4.10 2/2/18 Introduces piecewise cubic Hermite spline
// 4.03 11/11/17 Should be backward compatible for Igor 6.
// 4.02 Catch interpolate errors. 
// 4.01 Panel size decreased for compatibility with smaller or lower 
// resolution screens. Some bugs fixed.
// 4.00 Changed to panel interface. Major rewrite.
// 3.20 Removed option to prevent nodes from being inserted. Now the /NI 
// no insert flag is used by default, and we handle node insertion by 
// using a hook to collect ctrl-clicks.
// 3.10 Some changes to the global strings and variables in the package 
// folder. Tried and failed to add improved guesses for node positions.
// 3.01 11/4/16 Bug fix - waves with deltaX<0 produced nodes that were 
// reverse sorted in X, but Akima calculation cannot deal with reverse 
// sorted nodes. 
// 3.00 4/11/16 Added option for Akima style spline.
// 2.60 3/30/16 Added option to supress node addition. This is a toggle 
// that switches on the /NI option in GraphWaveEdit. Updated a lot of 
// other code. Changed some of the defaults (now saves nodes by default),
// so behaviour is not the same as previous version.
// 2.50 Log mode to save node positions to file. Can load node positions 
// from a file.
// 2.40 Added textbox reminder of how to toggle graph mode, renamed 
// procedures for consistency.
// 2.30 5/16/11 Bug fix - code was choking on non-existent wave in first 
// call to SBL_Clear(). Added panel to change fitting options.
// 2.10 Call SBL_Clear() from SBL_Subtract().
// 2.0 8/6/09 Incorporated suggestions from JJ Weimer that improve 
// functionality and clarity of code and compatibility with 6.10 ... and 
// then wrecked his nice programming with my clumsy code.
// 1.43b Allow refit function to switch interpolate flag when number of 
// nodes is too small to allow linear fit.
// 1.42 6/8/08 Set y axis to manual range to avoid crazy rescaling if 
// spline shoots out of range. Show subtracted spectrum whilst fitting.
// 1.41 10/9/07 For XY data we now interpolate directly into XY baseline.
// Improved guesses for initial node positions for unevenly spaced XY 
// data.
// 1.40 10/9/07 Fixed bug that occurred with baseline subtraction of XY 
// data.
// 1.30 7/24/07 Handles XY data. Baseline is waveform, but data wave need
// not be (though it should be reasonably closely spaced).
// 1.20 7/3/07 Uses new graphwaveedit flag. Now requires version 6.0.1.

Menu "Analysis"
	Submenu "Packages"
		"Spline Baseline",/Q, BaselineSpline#SBL_init()
	end
end

Menu "Macros"	
	"Spline Baseline",/Q, BaselineSpline#SBL_init()
end

// initialize spline baseline package
static function SBL_init()
	
	if (CheckForUpdates())
		return 0 // this file has been updated; quit to allow user to reload
	endif
	
	if (strlen(WinList("*",";","WIN:1"))==0)  // no graphs 
		doalert 0, "Spline baseline requires a trace plotted in a graph window."
		return 0
	endif
	
	// make sure the top graph is visible
	dowindow /F $WinName(0,1)
	
	// make sure graph has default axes
	getaxis /Q left
	if (v_flag==0)
		getaxis /Q bottom
	endif
	if (v_flag)
		doalert 0, "Trace must be plotted on bottom/left axes."
		return 0
	endif
	
	// clear any package detritus from the last used spline graph
	SBL_clear(1)
	
	// initialize (or reinitialize) the package data folder
	SBL_resetDFR()
	DFREF dfr = SBL_getDFREF()
	wave /SDFR=dfr w_spline_dependency
	
	SBL_makePanel()
		
	string graphStr=SBL_getGraph()
	controlinfo /W=$graphStr+"#SBL_panel" popTrace
	string traceStr=s_value
	
	variable success

	// place nodes within window
	success=SBL_resetNodes()
	if (!success)
		doalert 0, "Spline baseline requires a trace plotted in a graph window."
		SBL_clear(1)
		return 0
	endif
	
	// set the sub and base waves for this trace
	success=SBL_resetGraph(traceStr)
	if (!success)
		doalert 0, "Spline baseline requires a trace plotted in a graph window."
		SBL_Clear(1)
		return 0
	endif
	
	ModifyGraph zero(left)=1
	
	// set a dependency to trigger interpolation and update graph when we 
	// adjust a node position
	w_spline_dependency:=BaselineSpline#SBL_update(root:Packages:SplineFit:w_nodesY)	
	// won't work if we change package folder
	
	GraphWaveEdit /W=$graphStr /NI/T=1 /M $"w_nodesY"
	setWindow $graphStr hook(SBL_hook)=BaselineSpline#SBL_graphHook
		
	return 1	
end

static function SBL_graphHook(s)
	STRUCT WMWinHookStruct &s	
	
	if (s.eventcode!=5) // mouseup
		return 0
	endif
	
	variable v_key=GetKeyState(0) 
	if (v_key==0)
		return 0
	endif
	
	if(v_key&2^2) // shift click
		// toggle mode
		SBL_toggleEdit()
		return 0
	endif
	
	controlInfo /W=$s.winName+"#SBL_panel" check_edit
	if(v_value==0)
		return 0
	endif

	variable insert=0
	insert = (stringmatch(igorinfo(2),"Macintosh")) ? (v_key&2^4) : (v_key&2^0) // ctrl
	if (insert)
		DFREF dfr = SBL_getDFREF()
		wave /SDFR=dfr w_nodesY, w_nodesX
		variable v_X, v_Y
		v_X=AxisValFromPixel(s.winName, "bottom", s.mouseLoc.h)
		v_Y=AxisValFromPixel(s.winName, "left", s.mouseLoc.v)
		
		// deal with offset
		controlInfo /W=$s.winName+"#SBL_panel" popTrace
		string infoStr=traceinfo(s.winName,s_value,0)
		v_Y-=GetNumFromModifyStr(infoStr,"offset","{",1)
		
		w_nodesX[numpnts(w_nodesX)]={v_X}
		w_nodesY[numpnts(w_nodesY)]={v_Y}
		sort w_nodesX, w_nodesX, w_nodesY
		SBL_update({nan})
	endif

	return 0
end


static function SBL_toggleEdit()
	
	string graphStr=SBL_getGraph()
	controlinfo /W=$graphStr+"#SBL_panel" check_edit
	if (v_value)
		GraphNormal /W=$graphStr
		ModifyGraph /W=$graphStr mode(w_nodesY)=2
		checkbox check_edit win=$graphStr+"#SBL_panel", value=0
	else
		// manually scale y axis
		getaxis /W=$graphStr /Q left
		setAxis /W=$graphStr left, V_min, V_max	
		
		ModifyGraph /W=$graphStr mode(w_nodesY)=3
		GraphWaveEdit /W=$graphStr /NI/T=1 /M $"w_nodesY"	
		checkbox check_edit win=$graphStr+"#SBL_panel", value=1	
	endif
end

static function SBL_resetGraph(traceStr)
	string traceStr
	
	DFREF dfr = SBL_getDFREF()
	
	string graphStr=SBL_getGraph()
	wave /Z w_data=TraceNameToWaveRef(graphStr, traceStr)
	if (waveexists(w_data)==0)
		return 0
	endif
	wave /Z w_x=XWaveRefFromTrace(graphStr, traceStr)
	
	if (waveexists(w_x))
		// check for monotonic X wave
		make /free w_diff
		Differentiate w_x /D=w_diff
		Wavestats /Q/M=0 w_diff
		if (V_max>0 && V_min<0)
			doalert 0, "X wave is not monatonic! Quitting :("			
			SBL_clear(1)
			return 0
		endif	
	endif
	
	string xAxisName= StringByKey("XAXIS", traceinfo(graphStr,traceStr,0))
	string yAxisName= StringByKey("YAXIS", traceinfo(graphStr,traceStr,0))
		
	// manually scale y axis
	getaxis /Q left
	setAxis left, V_min, V_max	
		
	duplicate /O w_data dfr:w_base, dfr:w_sub
	wave /SDFR=dfr w_nodesX, w_nodesY, w_base, w_sub
	
	// make sure base and sub waves are displayed properly
	removefromgraph /Z /W=$graphStr w_base, w_sub
	if (waveexists(w_x))	
		appendtograph /W=$graphStr w_base vs w_x
		appendtograph /W=$graphStr w_sub vs w_x
	else
		appendtograph /W=$graphStr w_base
		appendtograph /W=$graphStr w_sub
	endif
	
	// if data are offset, then apply offsets to baseline
	string infoStr=traceinfo(graphStr,traceStr,0)
	variable trace_offset=GetNumFromModifyStr(infoStr,"offset","{",1)
	
	make /n=3 /free w_RGB={0,10000,65535}
	SBL_chooseColour(w_RGB, graphStr, traceStr)
	
	ModifyGraph /W=$graphStr offset(w_base)={0,trace_offset}
	ModifyGraph /W=$graphStr rgb(w_base)=(w_RGB[0],w_RGB[1],w_RGB[2])
	modifygraph /W=$graphStr live(w_base)=1	
	ModifyGraph /Z /W=$graphStr offset(w_nodesY)={0,trace_offset}
	ModifyGraph /W=$graphStr rgb(w_sub)=(0,0,0)
	ReorderTraces /W=$graphStr $"w_nodesY",{$"w_base"}

	return 1
end

static function SBL_resetNodes()
	
	string graphStr=SBL_getGraph()
	DFREF dfr = SBL_getDFREF()
	wave /SDFR=dfr w_nodesX, w_nodesY

	// define some local variables	
	variable i, p_low, p_high, delP, delX, numNodes=5
	string infoStr
	variable trace_offset=0
	
	controlinfo /W=$graphStr+"#SBL_panel" popTrace
	string traceStr=s_value
	
	wave /Z w_data=TraceNameToWaveRef(graphStr, traceStr)
	wave /Z w_x=XWaveRefFromTrace(graphStr, traceStr)
	if (waveexists(w_data)==0)
		return 0
	endif	
	redimension/N=(numNodes) w_nodesX, w_nodesY
	
	// determine where to place nodes on the bottom axis
	
	// keep nodes within window 
	GetAxis/Q bottom
	make /n=2 /free xrange={v_min, v_max}
	sort xrange, xrange
	
	GetAxis/Q left
	make /n=2 /free yrange={v_min, v_max}
	sort yrange, yrange
	
	// keep nodes within extent of w_data	
	if (waveexists(w_x))
		xrange[0]=max(xrange[0],wavemin(w_x))
		xrange[1]=min(xrange[1],wavemax(w_x))
	else
		xrange[0]=max(xrange[0], min(pnt2x(w_data, 0), pnt2x(w_data, numpnts(w_data)-1)) )
		xrange[1]=min(xrange[1], max(pnt2x(w_data, 0), pnt2x(w_data, numpnts(w_data)-1)) )
	endif

	delX=(xrange[1]-xrange[0])/(numNodes-1)
	// initialize node positions		
	w_nodesX=xrange[0]+p*delX
	
	for (i=0;i<numNodes;i+=1)
		if (waveexists(w_x))
			// necessary for unevenly spaced data
			findlevel /Q/P w_x, w_nodesX[i]-delX/24
			if (v_flag)
				findlevel /Q/P w_x, w_nodesX[i] // this HAS to be within range
			endif
			p_low = V_LevelX
			
			findlevel /Q/P w_x, w_nodesX[i]+delX/24
			if (v_flag)
				findlevel /Q/P w_x, w_nodesX[i] // this HAS to be within range
			endif
			p_high = V_LevelX
			wavestats /Q/M=1/R=[p_low , p_high] /Z  w_data
		else
			wavestats /Q/M=1/R=(w_nodesX[i]-delX/24 , w_nodesX[i]+delX/24) /Z  w_data
		endif
		w_nodesY[i]=V_avg  // this handles a few NANs	
	endfor
	
	// Akima calculation won't work if nodes are reverse sorted
	sort w_nodesX, w_nodesX, w_nodesY
	
	// if data are offset, then apply offsets to nodes
	infoStr=traceinfo(graphStr,traceStr,0)
	trace_offset=GetNumFromModifyStr(infoStr,"offset","{",1)
	
	// keep nodes within window
	w_nodesY=max(yrange[0]-trace_offset, w_nodesY)
	w_nodesY=min(yrange[1]-trace_offset, w_nodesY)
	
	checkdisplayed /W=$graphStr w_nodesY
	if (v_flag==0)
		appendtograph /W=$graphStr w_nodesY vs w_nodesX
	endif
	ModifyGraph /W=$graphStr mode(w_nodesY)=3,marker(w_nodesY)=11,rgb(w_nodesY)=(0,39168,0)
	ModifyGraph /W=$graphStr offset(w_nodesY)={0,trace_offset}, mode(w_nodesY)=3
		
	// manually scale y axis
	getaxis /W=$graphStr /Q left
	setAxis /W=$graphStr left, V_min, V_max
	
	GraphWaveEdit /W=$graphStr /NI/T=1 /M $"w_nodesY"
	Checkbox check_edit, win=$graphStr+"#SBL_panel", value=1
		
	return 1
end

// place nodes on w_data. doesn't plot waves.
// doesn't reset x positions unless they're out of range.
static function SBL_setNodes()
	
	string graphStr=SBL_getGraph()
	controlInfo /W=$graphStr+"#SBL_panel" popTrace
	string traceStr=s_value	
	wave /Z w_data=TraceNameToWaveRef(graphStr, traceStr)
	wave /Z w_x=XWaveRefFromTrace(graphStr, traceStr)
	
	DFREF dfr = SBL_getDFREF()
	wave /SDFR=dfr w_nodesX, w_nodesY
	variable numNodes=numpnts(w_nodesY)
	variable i, v_pnt, v_x
	
	for (i=0;i<numNodes;i+=1)
		if (waveexists(w_x))
			// necessary for unevenly spaced data
			v_x=max(wavemin(w_x), w_nodesX[i])
			v_x=min(wavemax(w_x), v_x)
			
			findlevel /Q/P w_x, v_x
			if (V_flag)
				v_pnt=numpnts(w_x)/2 // shouldn't arrive here
			else
				v_pnt=V_LevelX
			endif
			w_nodesY[i]=w_data[v_pnt]
		else					
			// keep x values within range of w_data		
			//	out of range error fixed with double precision nodes waves
			w_nodesX=max(min(pnt2x(w_data,0), pnt2x(w_data,numpnts(w_data)-1)), w_nodesX)
			w_nodesX=min(max(pnt2x(w_data,0), pnt2x(w_data,numpnts(w_data)-1)), w_nodesX)		
				
			w_nodesY[i]=w_data(w_nodesX[i]) 
		endif
	endfor

	#if IgorVersion() >= 7	
		// remove duplicate nodes, 
		// in case some were outside of the range of w_data
		duplicate /free w_nodesX w_duplicates
		findduplicates /SN=(nan) /SNDS=w_duplicates  w_nodesX
		i=0
		do
			if(numtype(w_duplicates[i])!=0)
				deletepoints i, 1, w_nodesX, w_nodesY, w_duplicates
			else
				i+=1
			endif
		while(i<numpnts(w_nodesX))
	#else
		// you're out of luck
	#endif
	
	if(numpnts(w_nodesX)<2)
		SBL_resetNodes()
		doalert 0, "Nodes couldn't be placed on "+nameofwave(w_data)+" at the same x positions. Nodes have been reset."
	endif
	
end

static function SBL_makePanel()
		
	DFREF dfr = SBL_getDFREF()
	
	// make sure the top graph is visible
	string graphStr=WinName(0,1)
	dowindow /F $graphStr
	
	// make sure there's space to the left of the graph for the panel
	getwindow /Z $graphStr wsize
	if (V_left<200)
	#if IgorVersion() >= 7	
		movewindow /W=$graphStr 200, V_top, -1, -1
	#else
		movewindow /W=$graphStr 200, V_top, 200+(V_right-V_left), v_bottom
	#endif
	endif

	// make panel
	NewPanel /K=1/N=SBL_panel/W=(200,0,0,380)/HOST=$graphStr/EXT=1 as "Spline Controls"	
	ModifyPanel /W=$graphStr#SBL_panel, noEdit=1		
	
	string traceStr=stringfromlist(0, SBL_traces())
	wave /Z w=TraceNameToWaveRef(graphStr, traceStr)
	
	variable i=0, deltaY=25, font=12, groupw=180
	v_left=30
	v_top=5

	GroupBox group0,pos={v_left-20,v_top+deltaY*i},size={groupw,deltaY*2},title="Data wave"
	GroupBox group0,fSize=font
	i+=1	
	// store values internally in these controls
	PopupMenu popTrace, mode=1, Value=BaselineSpline#SBL_traces(), title="",pos={v_left,v_top+deltaY*i},size={130,20}
	PopupMenu popTrace, help={"select data wave" }, proc=BaselineSpline#SBL_popup, fSize=font
	i+=1.5	
	
	GroupBox group1 pos={v_left-20,v_top+deltaY*i},size={groupw,deltaY*6.0},title="Spline parameters"
	GroupBox group1 fSize=font
	i+=1	
	Checkbox check_akima, pos={v_left,v_top+deltaY*i}, fSize=font, value=0, proc=BaselineSpline#SBL_checkboxes, title="Akima"
	Checkbox check_akima, mode=1,help={"use Akima spline"}	
	i+=0.8
	Checkbox check_PCHIP, pos={v_left,v_top+deltaY*i}, fSize=font, value=0, proc=BaselineSpline#SBL_checkboxes, title="PCHIP"
	Checkbox check_PCHIP, mode=1,help={"use piecewise cubic Hermite spline"}	
	i+=0.8
	Checkbox check_cubic, pos={v_left,v_top+deltaY*i}, fSize=font, value=1, proc=BaselineSpline#SBL_checkboxes, title="Cubic"
	Checkbox check_cubic, mode=1,help={"use cubic spline"}
	i+=0.8
	PopupMenu popE,pos={v_left+35,v_top+deltaY*i},size={46,21}, fSize=font, title="Ends:", proc=BaselineSpline#SBL_popup
	PopupMenu popE, value="1st deriv.;natural", mode=2
	i+=0.8
	Checkbox check_linear, pos={v_left,v_top+deltaY*i}, fSize=font, value=0, proc=BaselineSpline#SBL_checkboxes, title="Linear"
	Checkbox check_linear, mode=1,help={"lines between nodes"}
	i+=0.8
	Checkbox check_smoothing, pos={v_left,v_top+deltaY*i}, fSize=font, value=0, proc=BaselineSpline#SBL_checkboxes, title="Smoothing"
	Checkbox check_smoothing, mode=1,help={"use smoothing spline"}
	SetVariable setvarF, pos={v_left+100,v_top+deltaY*i},size={50,16},title="",bodyWidth=50
	SetVariable setvarF, value=_NUM:1, limits={0,inf,1}, fSize=font, proc=BaselineSpline#SBL_setvar
	i+=1.5
	
	GroupBox group2,pos={v_left-20,v_top+deltaY*i},size={groupw,deltaY*4.5},title="Node control"
	GroupBox group2,fSize=font
	i+=1	
	Checkbox check_edit, pos={v_left+30,v_top+deltaY*i}, fSize=font, value=1, proc=BaselineSpline#SBL_Checkboxes, title="Edit mode "
	Checkbox check_edit, mode=0,help={"Toggle to edit nodes"}, side=1
	i+=0.8
	SetDrawLayer ProgBack
	SetDrawEnv textyjust=2
	DrawText v_left-12,v_top+deltaY*i,"Shift-click to toggle mode, ctrl-\r/option-click to add/zap nodes"
	i+=1.5
	Button SBL_nodes,pos={v_left,v_top+deltaY*i},size={50,20},title="Reset", proc=BaselineSpline#SBL_Buttons
	Button SBL_nodes,help={"Distribute nodes over x-range of graph"}, fSize=font
	
	Button SBL_load,pos={v_left+80,v_top+deltaY*i},size={50,20},title="Load...", proc=BaselineSpline#SBL_Buttons
	Button SBL_load,help={"Load nodes used in a previous fit"} , fSize=font
	
	i+=1.5
	Button SBL_sub,pos={65,v_top+deltaY*i},size={70,20},title="Subtract", proc=BaselineSpline#SBL_Buttons
	Button SBL_sub,help={"Subtract baseline and save a copy"}, fSize=font
	
	return 1
end

static function SBL_setvar(s)
	STRUCT WMSetVariableAction &s
	
	if(s.eventCode==-1)
		return 0
	endif
	SBL_Update({nan})	
end

static function SBL_popup(s)
	STRUCT WMPopupAction &s
	
	if(s.eventCode==-1)
		return 0
	endif
	
	if (stringmatch(s.ctrlName, "popTrace"))
		if ( SBL_ResetGraph(s.popStr) )
			SBL_setNodes()
			SBL_Update({nan})
		endif
		
		//controlInfo /W=$s.win+"#SBL_panel" check_edit
		controlInfo /W=$s.win check_edit
		
		// not sure why this is needed, there's no reason why it should get out of sync
		if (v_value)
			string graphStr=parseFilepath(0, s.win, "#", 0, 0)
			GraphWaveEdit /W=$graphStr /NI/T=1 /M $"w_nodesY"
		endif
	else
		SBL_Update({nan})	
	endif	
end

static function SBL_checkboxes(s)
	STRUCT WMCheckboxAction &s
	
	if(s.eventCode==-1)			
		SBL_clear(0)
		return 0
	endif
	
	string buttonsList="check_akima;check_PCHIP;check_cubic;check_linear;check_smoothing"
	variable i
	variable numTypes=itemsinlist(buttonsList)
	variable buttonNum=WhichListItem(s.ctrlName, buttonsList)

	if(buttonNum>=0)
		for (i=0;i<numTypes;i+=1)
			if(i!=buttonNum)
				Checkbox $StringFromList(i, buttonsList), win=$s.win, value=0
			endif
		endfor
		SBL_Update({nan})
		return 1
	endif
	
	// only remaining checkbox is check_edit
	string graphStr=parseFilepath(0, s.win, "#", 0, 0)
	if (s.checked)
		// manually scale y axis
		getaxis /W=$graphStr /Q left
		setAxis /W=$graphStr left, V_min, V_max
		
		ModifyGraph /W=$graphStr mode(w_nodesY)=3
		GraphWaveEdit /W=$graphStr /NI/T=1 /M $"w_nodesY"
	else
		GraphNormal /W=$graphStr
		ModifyGraph /W=$graphStr mode(w_nodesY)=2		
	endif

	return 1
end

static function SBL_buttons(s)
	STRUCT WMButtonAction &s
	
	if(s.eventCode!=2)			
		return 0
	endif
	
	if (stringmatch(s.ctrlname, "SBL_sub"))
		SBL_subtract()
	endif
	
	if (stringmatch(s.ctrlname, "SBL_nodes"))
		SBL_resetNodes()
	endif
	
	if (stringmatch(s.ctrlname, "SBL_load"))
		SBL_loadNodes()
	endif
	return 0
end

static function /S SBL_traces()
	
	string graphStr=SBL_getGraph()
	string listStr=TraceNameList(graphStr,";",1+4)
	string removeStr="w_base;w_NodesY;"
	removeStr+=listmatch(listStr, "*_sub")
	removeStr+=listmatch(listStr, "*_BL")
	listStr=RemoveFromList(removeStr, listStr, ";", 0)
	return listStr
end

// initialize (or reinitialize) the package data folder
static function /DF SBL_resetDFR()
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:SplineFit	
	DFREF dfr = root:Packages:SplineFit	

	wave /Z w_nodesX=dfr:w_nodesX
	if (WaveType(w_nodesX)!=4)	
		make /D/O/n=5 dfr:w_nodesX, dfr:w_nodesY
	endif
	make /O/n=1  dfr:w_spline_dependency={nan}	
	return dfr
end

static function/DF SBL_getDFREF()
	DFREF dfr = root:Packages:SplineFit
	if (DataFolderRefStatus(dfr) != 1)	
		DFREF dfr = SBL_resetDFR()
	endif
	return dfr
end

// dependency function to define the spline curve
static function SBL_update(w)
	wave w
	
	DFREF dfr = SBL_getDFREF()
	wave /Z/SDFR=dfr w_nodesX,w_nodesY,w_base,w_sub
	
	string graphStr=SBL_getGraph()
	if (strlen(graphStr)==0)
		return 0
	endif
		
	controlInfo /W=$graphStr+"#SBL_panel" popTrace
	string traceStr=s_value
	wave /Z w_data=TraceNameToWaveRef(graphStr, traceStr)
	wave /Z w_x=XWaveRefFromTrace(graphStr, traceStr)
	
	if (waveexists(w_data)==0)
		return nan
	endif
	
	if(numpnts(w_nodesX)<2)
		return nan
	endif
		
	if (numpnts(w_data)!=numpnts(w_base))
		// could get here if we trigger an update before popup runs updateGraph
		return nan
	endif
	
	// flags for interpolate2
	variable flagE, flagF, type=1 // set default to linear interpolation
	// if type < 1, spline is calculated using a method other than Igor's interpolate2
	
	controlInfo /W=$graphStr+"#SBL_panel" check_akima
	if (v_value)
		type = 0
	endif
	
	controlInfo /W=$graphStr+"#SBL_panel" check_PCHIP
	if (v_value)
		type = -1
	endif
	
	if (type==0 && numpnts(w_nodesX)<5)
		doalert 0, "Akima spline requires at least 5 nodes\r - reverting to standard cubic spline"
		Checkbox check_akima, win=$graphStr+"#SBL_panel", value=0	
		Checkbox check_cubic, win=$graphStr+"#SBL_panel", value=1
		type=2
	endif
		
	// check for small number of points
	if (type>0 && numpnts(w_nodesX)<4)
		type=1 // don't uncheck spline type, so will revert to desired type as nodes are added
	endif
		
	if (type>0) // types that use interpolate2
		do
			// collect info to set the flags for interpolate2
			controlInfo /W=$graphStr+"#SBL_panel" popE
			flagE=V_Value
			controlInfo /W=$graphStr+"#SBL_panel" check_cubic
			if (v_value)
				type=2
				break
			endif
			controlInfo /W=$graphStr+"#SBL_panel" check_smoothing
			if(v_value)
				type=3
				controlInfo /W=$graphStr+"#SBL_panel" setvarF
				flagF=V_value
				break
			endif
		while(0)
	endif

	// do the spline calculation 
	if (type==-1) // PCHIP
		SBL_PCHIP(w_nodesX, w_nodesY, w_base, xwave=w_x)
	elseif (type==0) // Akima type spline
		SBL_Akima(w_nodesX, w_nodesY, w_base, xwave=w_x)	
	elseif(type > 0) // cubic spline
		DebuggerOptions
		variable sav_debug=V_debugOnError 
		DebuggerOptions debugOnError=0 // switch this off in case interpolate fails
		try
			if(waveexists(w_x))
				Interpolate2 /E=(flagE) /F=(flagF) /T=(type) /I=3  /X=w_x /Y=w_base w_nodesX, w_nodesY; AbortOnRTE
			else
				Interpolate2 /E=(flagE) /F=(flagF) /T=(type) /I=3  /Y=w_base  w_nodesX, w_nodesY; AbortOnRTE
			endif
		catch
			Print "Error during interpolate:"
			Variable CFerror = GetRTError(1)
			Print GetErrMessage(CFerror)
		endtry
		DebuggerOptions debugOnError=sav_debug
	endif	
	
	w_sub=w_data-w_base
	
	return nan	
end

// function to subtract the spline curve
static function SBL_subtract()
	
	variable appendBL=0
	
	string graphStr=SBL_getGraph()
	controlInfo /W=$graphStr+"#SBL_panel" popTrace
	string traceStr=s_value
	wave /Z w_data=TraceNameToWaveRef(graphStr, traceStr)
	wave /Z w_x=XWaveRefFromTrace(graphStr, traceStr)
	
	if (waveexists(w_data)==0)
		return nan
	endif
	
	DFREF dfr = SBL_getDFREF()
	wave /SDFR=dfr w_nodesX,w_nodesY,w_base
	
	string strSubtracted, strBaseline, strNodes
	strSubtracted=CleanupName( nameofwave(w_data)+"_sub",0)
	strBaseline=CleanupName( nameofwave(w_data)+"_BL",0)
	strNodes=CleanupName( nameofwave(w_data)+"_SpNodes",0)
	
	DFREF currentDFR=GetDataFolderDFR()
	
	if (exists(strSubtracted))
		doalert 1, strSubtracted+" exists. Overwrite?"
		if(V_flag==2)
			return 0
		endif
	endif
	duplicate /o w_data currentDFR:$strSubtracted /WAVE=nob
	
	if (exists(strBaseline))
		doalert 1, strBaseline+" exists. Overwrite?"
		if(V_flag==2)
			return 0
		endif
	endif
	duplicate /o w_base currentDFR:$strBaseline /WAVE=bl
	
	if (exists(strNodes))
		doalert 1, strNodes+" exists. Overwrite?"
		if(V_flag==2)
			return 0
		endif
	endif
	duplicate /o w_nodesX currentDFR:$strNodes /WAVE=nd
	redimension /N=(-1, 2) nd
	nd[][0]=w_nodesX[p]
	nd[][1]=w_nodesY[p]
	
	nob=w_data-bl
	
	string noteStr="Baseline Parameters\r"
	controlInfo /W=$graphStr+"#SBL_panel" check_akima
	if (v_value)
		noteStr+="type=AkimaSpline;"
	endif
	controlInfo /W=$graphStr+"#SBL_panel" check_cubic
	if (v_value)
		noteStr+="type=CubicSpline;"
		controlInfo /W=$graphStr+"#SBL_panel" popE
		noteStr+="Ends="+s_value+";"
	endif
	controlInfo /W=$graphStr+"#SBL_panel" check_smoothing
	if(v_value)
		noteStr+="type=SmoothingSpline;"
		controlInfo /W=$graphStr+"#SBL_panel" setvarF
		noteStr+="SmoothingFactor="+num2str(v_value)+";"
	endif
	controlInfo /W=$graphStr+"#SBL_panel" check_linear
	if(v_value)
		noteStr+="type=LinearSpline;"
	endif
	noteStr+="numNodes="+num2str(numpnts(w_nodesX))+";"
	noteStr+="data="+nameofwave(w_data)+";output="+nameofwave(nob)+","+nameofwave(bl)+","+nameofwave(nd)+";\r"
	note /k bl, noteStr
	note nob, noteStr // append note from baseline wave to output wave note
	printf noteStr
		
	// add baseline-subtracted output wave to plot
	checkdisplayed /W=$graphStr nob
	if(v_flag==0)
		if(waveexists(w_x))
			appendtograph /W=$graphStr /C=(0,0,0) nob vs w_x
		else
			appendtograph /W=$graphStr /C=(0,0,0) nob
		endif
	endif
	
	// possibly plot baseline
	if(appendBL)
		checkdisplayed /W=$graphStr bl
		if(v_flag==0)
			if(waveexists(w_x))
				appendtograph /W=$graphStr /C=(0,0,0) bl vs w_x
			else
				appendtograph /W=$graphStr /C=(0,0,0) bl
			endif
		endif
	endif	
	
	return 1
end

// function to clear the spline curve
static function SBL_clear(killPanel)
	variable killPanel
	
	string graphStr=SBL_getGraph()
	
	if (strlen(graphStr))
		GraphNormal /W=$graphStr
		RemoveFromGraph /W=$graphStr/Z w_base,w_nodesY,w_sub
		setWindow $graphStr hook(SBL_hook)=$""
	endif
	
	DFREF dfr = SBL_getDFREF()
	wave /Z w=dfr:w_spline_dependency
	if (waveexists(w))
		SetFormula w, "" // remove dependency
	endif
	
	if(killPanel && strlen(graphStr))
		killwindow $graphStr+"#SBL_panel"
	endif

	return 1
end

static function SBL_loadNodes()
	DFREF dfr = SBL_getDFREF()
			
	string str_NodeWave
	variable LoadY
	
	prompt str_NodeWave, "Select saved nodes", popup, wavelist("*",";","DIMS:2") 
	prompt LoadY, " ", popup, "Load X positions only;Load X and Y positions;"
	DoPrompt "", str_NodeWave, LoadY
	if (V_Flag)
		return 0
	endif
	
	wave w_nodes=$str_NodeWave
	loadY-=1
	
	wave /SDFR=dfr w_nodesX, w_nodesY
	
	redimension /N=(dimsize(w_nodes, 0)), w_nodesX, w_nodesY
	w_nodesX=w_nodes[p][0]
	
	if (LoadY)
		w_nodesY=w_nodes[p][1]
		return 1
	endif

	SBL_setNodes()

	return 1
end

// set colour wave w_color for visibility against w_target.
static function SBL_chooseColour(w_color, graphStr, traceStr)
	wave w_color
	string graphStr, traceStr
	
	// figure out a contrasting colour for baseline and sub waves
	variable c0,c1,c2
	string infoStr=traceinfo(graphStr, traceStr, 0)
	variable start=strsearch(infoStr, ";rgb(x)=", 0)
	sscanf infoStr[start,strlen(infoStr)-1], ";rgb(x)=(%d,%d,%d*", c0,c1,c2
	make /free w_cTrace={c0,c1,c2}, w_index={1,2,3}
	MakeIndex /R w_cTrace, w_index
	IndexSort  w_index, w_color 
	return 1	
end

// ---------------------- Akima spline code --------------------------

// this part based on code posted to IgorExchange by Michael Bongard
// edited by tony.withers@uwo.ca to allow extrapolation beyond end nodes
 
// Akima.ipf: Routines to implement Akima-spline fitting, based on
// H. Akima, JACM, Vol 17, No 4, 1970 p 589-602
// after M. Bongard, 11/17/09

static function /wave SBL_AkimaIota(knotX, knotY)
	wave knotX, knotY
 	
 	// removed some sanity checks, converted to wave function
	Variable numKnots = numpnts(knotX)
 
	if(numKnots < 5)
		Print "calcIota: ERROR -- Akima spline algorithm requires at least 5 knots. Aborting..."
		return $""
	endif
 	 
	Make/D/FREE/N=(numKnots)  kX, kY
	kX = knotX
	kY = knotY
 
	// Handle end-point extrapolation
	Make/D/FREE/N=5 endX, endY
	
	// RHS: end points are last three in dataset
	endX[0,2] = kX[p+numKnots-3]
	endY[0,2] = kY[p+numKnots-3]
 
	SBL_AkimaEndPoints(endX, endY)
 	
 	kX[numpnts(kX)]={endX[3],endX[4]}
 	kY[numpnts(kY)]={endY[3],endY[4]}

	// LHS: end points are first three in dataset, but reversed in ordering
	// (i.e. point 3 in Akima's notation == index 0)
	endX = 0
	endY = 0
	
	endX[0,2] = knotX[2-p]
	endY[0,2] = knotY[2-p]
 
	SBL_AkimaEndPoints(endX, endY)
	
	InsertPoints  0, 2, kX, kY
	kX[0,1] = endX[4-p]
	kY[0,1] = endY[4-p]
 
	// kX, kY are now properly populated, including extrapolated endpoints
 
	Make/D/FREE/N=(numKnots + 4 - 1)  mK
	mK = (kY[p+1]-kY)/(kX[p+1]-kX)	
 
	Make/free/N=(numKnots) w_Iota
	w_Iota= ( abs(mK[p+3]-mK[p+2])*mK[p+1] + abs(mK[p+1]-mK[p])*mK[p+2] ) / (abs(mK[p+3]-mK[p+2])+abs(mK[p+1]-mK))
	w_Iota= numtype(w_Iota)==0 ? w_Iota : 0.5*(mK[p+1]+mK[p+2])
	return w_Iota
end 

 
// Given: 5-point knot wave knotX, knotY, with i=[0,2] representing the last three
// knot locations from data, compute end knots i=[3,4] appropriately.
threadSafe static function SBL_AkimaEndPoints(kX, kY)
	WAVE kX, kY // node X,Y coordinate locations, respectively
 
	// compute X locations of knots, eq. 8 of Akima 1970:
	kX[3] = kX[1] + kX[2] - kX[0]
	kX[4] = 2*kX[2] - kX[0]
 
	// eq. (12)-(14), determine gradient on [0,1]
	Make/N=4/FREE  mi	
	mi = (kY[p+1]-kY)/(kX[p+1]-kX)
 
	// Determine remainder of quantities by applying solutions of eq (9)
	kY[3] = (2*mi[1] - mi[0])*(kX[3] - kX[2]) + kY[2]
	mi[2] = (kY[3] - kY[2]) / (kX[3] - kX[2])
 
	kY[4] = (2*mi[2] - mi[1])*(kX[4] - kX[3]) + kY[3]
	// not sure why the next line was included
	//mi[3] = (kY[4] - kY[3]) / (kX[4] - kX[3])
end


static function SBL_Akima(w_nodesX, w_nodesY, w_interp [, xwave])
	wave w_nodesX, w_nodesY, w_interp
	wave/Z xwave
	
	wave w_iota=SBL_AkimaIota(w_nodesX, w_nodesY)
	if (waveexists(w_iota))
		if(waveexists(xwave))
			MultiThread w_interp=SBL_AkimaThread(xwave[p], w_nodesX, w_nodesY, w_iota)
		else
			MultiThread w_interp=SBL_AkimaThread(x, w_nodesX, w_nodesY, w_iota)
		endif
		return 1
	endif	
	return 0
end


threadSafe static function SBL_AkimaThread(x, w_nodesX, w_nodesY, w_Iota)
	Variable x
	WAVE w_nodesX, w_nodesY, w_Iota
 	
	variable p0, p1
	if(x<wavemin(w_nodesX))
		p0=0
	elseif(x>=wavemax(w_nodesX))
		p0=numpnts(w_nodesX)-2
	else
		findlevel /P/EDGE=1/Q w_nodesX, x
		if(v_flag==1) // not found			
			return nan
		endif
		p0=floor(v_levelX)
	endif	
	p1=p0+1

	Variable x1, x2, y1, y2, iota1, iota2
	x1 = w_nodesX[p0]
	y1 = w_nodesY[p0]
	iota1 = w_Iota[p0]
 
	x2 = w_nodesX[p1]
	y2 = w_nodesY[p1]
	iota2 = w_Iota[p1]
 
	Variable coeff2, coeff3, delx
	coeff2 = ( 3 * (y2 - y1)/(x2 - x1) - 2*iota1 - iota2 ) / (x2 - x1)
	coeff3 = (iota1 + iota2 - 2 * (y2 - y1)/(x2 - x1)) / (x2 - x1)^2
	delx = x - x1
 
	return y1 + iota1*delx + coeff2*delx^2 + coeff3*delx^3
end

// ----------------------------- PCHIP ---------------------------

// populate w_interp with interpolated values from piecewise cubic 
// Hermite spline
static function SBL_PCHIP(w_nodesX, w_nodesY, w_interp, [xwave])	
	wave w_nodesX, w_nodesY, w_interp
	wave /Z xwave
	
	if(numpnts(w_nodesX)==2) // linear interpolation
		if(waveexists(xwave))	
			w_interp=w_nodesY[0]+(xwave[p]-w_nodesX[0])/(w_nodesX[1]-w_nodesX[0])*(w_nodesY[1]-w_nodesY[0])
		else
			w_interp=w_nodesY[0]+(x-w_nodesX[0])/(w_nodesX[1]-w_nodesX[0])*(w_nodesY[1]-w_nodesY[0])
		endif
		return 1
	endif
	
	wave w_d=SBL_setDerivativesPCHIP(w_nodesX, w_nodesY)	
	if(waveexists(xwave))	
		w_interp=SBL_InterpolatePCHIP(w_nodesX, w_nodesY, w_d, xwave[p])
	else
		w_interp=SBL_InterpolatePCHIP(w_nodesX, w_nodesY, w_d, x)
	endif	
end

// The trick is to choose good values for the slope at each node. See:
// FN Fritsch and RE Carlson (1980) Monotone piecewise cubic 
// interpolation. SIAM J. Numer. Anal. 17: 238. DOI:10.1137/0717021 
// KW Brodlie (1980) A review of methods for curve and function drawing. 
// In Mathematical Methods in Computer Graphics and Design, KW Brodlie, 
// ed., Academic Press, London, pp. 1-37.
// FN Fritsch and J Butland (1984) A method for constructing local 
// monotone piecewise cubic interpolants, SIAM Journal on Scientific and 
// Statistical Computing 5: 300-304.
static function /WAVE SBL_setDerivativesPCHIP(w_nodesX, w_nodesY)
	wave w_nodesX, w_nodesY

	duplicate /free w_nodesX, w_d, w_a
	// w_d will be set to desired derivatives at node positions
	// w_a will be used as weighting for harmonic means
	w_d=0
	
	make /free/n=(numpnts(w_nodesX)-1), w_m, w_delx
	// w_m[i] is gradient between node i and i+1
	// w_delx[i] and w_delx[i-1] are x offsets of nodes i+1 and i-1	
	
	w_m = (w_nodesY[p+1]-w_nodesY)/(w_nodesX[p+1]-w_nodesX)
	w_delx = w_nodesX[p+1]-w_nodesX
	variable pmax=numpnts(w_d)-2  // pmax is numpnts(w_m)-1 = numpnts(w_d)-2
	
	w_a[1,pmax] = (1+(w_delx[p-1])/(w_delx[p-1]+w_delx))/3	
	
	// Brodlie modification of Butland formula (for same sign slopes)	
	w_d [1,pmax] = ( w_m!=0 && w_m[p-1]!=0 && (sign(w_m)==sign(w_m[p-1])) ) ? w_m[p-1]*w_m/(w_a*w_m[p-1]+(1-w_a)*w_m) : 0
		
	// deal with ends
	w_d[0]=SBL_EndDerivativePCHIP(w_delx[0], w_delx[1], w_m[0], w_m[1])
	w_d[pmax+1]=SBL_EndDerivativePCHIP(w_delx[pmax], w_delx[pmax-1], w_m[pmax], w_m[pmax-1])
	
	// clean up any division by zero errors
	w_d = numtype(w_d)==0 ? w_d : 0
			
	return w_d
end

// one-sided three-point estimate for the derivative adjusted to be shape
// preserving
static function SBL_EndDerivativePCHIP(delX0, delX1, m0, m1)
	variable delX0, delX1, m0, m1
	
	variable derivative
	
	derivative = ( m0*(2*delX0+delX1) - delX0*m1 ) / (delX0+delX1)
	if(sign(derivative)!=sign(m0))
		derivative=0
	elseif( (sign(m0)!=sign(m1)) && (abs(derivative)>3*abs(m0)) )
		derivative=3*m0
	endif
	return derivative
end

threadsafe static function SBL_InterpolatePCHIP(w_nodesX, w_nodesY, w_nodesM, x)
	wave w_nodesX, w_nodesY, w_nodesM
	variable x
		
	variable p0, p1
	if(x<wavemin(w_nodesX))
		p0=0
	elseif(x>=wavemax(w_nodesX))
		p0=numpnts(w_nodesX)-2
	else
		findlevel /P/EDGE=1/Q w_nodesX, x
		if(v_flag==1) // not found			
			return nan
		endif
		p0=floor(v_levelX)
	endif
	
	p1=p0+1		
	return SBL_InterpolateHermite(x, w_nodesX[p0], w_nodesY[p0], w_nodesX[p1], w_nodesY[p1], w_nodesM[p0], w_nodesM[p1])
end

// https: en.wikipedia.org/wiki/Cubic_Hermite_spline
// calculate cubic function f(x) such that f(x1)=y1, f(x2)=y2, f'(x1)=m1 and f'(x2)=m2
threadsafe static function SBL_InterpolateHermite(x, x0, y0, x1, y1, m0, m1)
	variable x, x0, y0, x1, y1, m0, m1
	
	variable t=(x-x0)/(x1-x0)
	m0*=(x1-x0)
	m1*=(x1-x0)
	return y0*(2*t^3-3*t^2+1) + m0*(t^3-2*t^2+t) + y1*(-2*t^3+3*t^2) + m1*(t^3-t^2)	
end

// ----------------------------------------------------------------------------


// for Igor 6 compatibility
static function /T SBL_getGraph()

	#if IgorVersion() >= 7	
		GetWindow /Z SBL_panel  activeSW
		return parseFilepath(0, s_value, "#", 0, 0)
	#endif	
	
	string listStr=WinList("*", ";","WIN:1")
	string panelStr, graphStr
	variable i
	do
		graphStr=stringfromlist(i, listStr)
		panelStr=graphStr+"#SBL_panel"
		if (wintype(panelStr))
			return graphStr
		endif
		i+=1
	while (i<itemsinlist(listStr))
	return ""
end

// execute BaselineSpline#makeSpectrum() to create a fake spectrum for demo
// creates a fake 'spectrum' containing Gaussian peaks 
// superimposed on a noisy baseline
static function makeSpectrum([n])
	variable n
		
	n=ParamIsDefault(n)?1:n
	
	variable numPeaks=5 // number of Gaussian peaks in the spectrum
	string strName
	variable a,b,c,d,i,j,height,position,width
	
	display /K=1
	for(i=0;i<n;i+=1)
		strName=uniquename("demo", 1, 0)
		make /n=1000 $(uniquename("demo", 1, 0)) /WAVE=demo
		a=enoise(5); b=enoise(1e-3); c=enoise(1e-5); d=enoise(400)
		demo=gnoise(0.1)+a+b*x+c*(x-d)^2
		for(j=0;j<numPeaks;j+=1)
			height=150+enoise(50)
			position=500+enoise(400)
			width=20+enoise(10)
			demo+=height*gauss(x, position, width/sqrt(2))
		endfor
		appendtograph demo
	endfor
end